import type { PeerId } from "@libp2p/interface-peer-id";
import type { Source, Target, Node, Bound, KeyValueStore } from "@canvas-js/okra";
import { EventEmitter } from "@libp2p/interface/events";
import { Logger } from "@libp2p/logger";
import type { Message } from "@canvas-js/interfaces";
import { Signature } from "@canvas-js/signed-cid";
import { Awaitable } from "./utils.js";
export interface ReadOnlyTransaction {
    messages: Omit<KeyValueStore, "set" | "delete"> & Source;
    parents: Omit<KeyValueStore, "set" | "delete">;
}
export interface ReadWriteTransaction {
    messages: KeyValueStore & Target;
    parents: KeyValueStore;
}
export interface MessageLogInit<Payload = unknown, Result = void> {
    topic: string;
    apply: (id: string, signature: Signature | null, message: Message<Payload>) => Awaitable<Result>;
    validate: (payload: unknown) => payload is Payload;
    signatures?: boolean;
    sequencing?: boolean;
    replay?: boolean;
}
export interface MessageSigner<Payload = unknown> {
    sign: (message: Message<Payload>) => Awaitable<Signature | null>;
}
export type MessageLogEvents<Payload, Result> = {
    message: CustomEvent<{
        id: string;
        signature: Signature | null;
        message: Message<Payload>;
        result: Result;
    }>;
    commit: CustomEvent<{
        topic: string;
        root: Node;
    }>;
    sync: CustomEvent<{
        topic: string;
        peerId: PeerId;
    }>;
};
export declare abstract class AbstractMessageLog<Payload = unknown, Result = unknown> extends EventEmitter<MessageLogEvents<Payload, Result>> {
    #private;
    private static defaultSigner;
    private static bound;
    abstract close(): Promise<void>;
    protected abstract entries(lowerBound?: Bound<Uint8Array> | null, upperBound?: Bound<Uint8Array> | null, options?: {
        reverse?: boolean;
    }): AsyncIterable<[key: Uint8Array, value: Uint8Array]>;
    protected abstract read<T>(callback: (txn: ReadOnlyTransaction) => Awaitable<T>, options?: {
        target?: PeerId;
    }): Promise<T>;
    protected abstract write<T>(callback: (txn: ReadWriteTransaction) => Awaitable<T>, options?: {
        source?: PeerId;
    }): Promise<T>;
    readonly topic: string;
    readonly apply: (id: string, signature: Signature | null, message: Message<Payload>) => Awaitable<Result>;
    readonly validate: (payload: unknown) => payload is Payload;
    readonly signatures: boolean;
    readonly sequencing: boolean;
    protected readonly log: Logger;
    protected readonly mempool: Mempool<Payload>;
    protected constructor(init: MessageLogInit<Payload, Result>);
    protected replay(): Promise<void>;
    iterate(lowerBound?: {
        id: string;
        inclusive: boolean;
    } | null, upperBound?: {
        id: string;
        inclusive: boolean;
    } | null, options?: {
        reverse?: boolean;
    }): AsyncIterable<[id: string, signature: Signature | null, message: Message<Payload>]>;
    encode(signature: Signature | null, message: Message<Payload>): [key: Uint8Array, value: Uint8Array];
    decode(value: Uint8Array): [id: string, signature: Signature | null, message: Message<Payload>];
    getClock(): Promise<[clock: number, parents: string[]]>;
    get(id: string): Promise<[signature: Signature | null, message: Message<Payload> | null]>;
    private getParents;
    append(payload: Payload, options?: {
        signer?: MessageSigner<Payload>;
    }): Promise<{
        id: string;
        signature: Signature | null;
        message: Message<Payload>;
        result: Result;
    }>;
    insert(signature: Signature | null, message: Message<Payload>): Promise<{
        id: string;
    }>;
    sync(sourcePeerId: PeerId, source: Source): Promise<{
        root: Node;
    }>;
    serve(targetPeerId: PeerId, callback: (source: Source) => Promise<void>): Promise<void>;
}
declare class Mempool<Payload> {
    readonly messages: Map<string, {
        signature: Signature | null;
        message: Message<Payload>;
    }>;
    readonly dependencies: Map<string, Set<string>>;
    readonly children: Map<string, Set<string>>;
}
export {};
