var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _AbstractMessageLog_instances, _AbstractMessageLog_apply;
import { CustomEvent, EventEmitter } from "@libp2p/interface/events";
import { logger } from "@libp2p/logger";
import { equals } from "uint8arrays";
import { bytesToHex as hex } from "@noble/hashes/utils";
import { verifySignature } from "@canvas-js/signed-cid";
import { decodeId, encodeId, decodeSignedMessage, encodeSignedMessage, getClock } from "./schema.js";
import { Driver } from "./sync/driver.js";
import { assert, nsidPattern, cborNull } from "./utils.js";
export class AbstractMessageLog extends EventEmitter {
    constructor(init) {
        super();
        _AbstractMessageLog_instances.add(this);
        this.mempool = new Mempool();
        assert(nsidPattern.test(init.topic), "invalid topic (must match [a-zA-Z0-9\\.\\-]+)");
        this.topic = init.topic;
        this.apply = init.apply;
        this.validate = init.validate;
        this.signatures = init.signatures ?? true;
        this.sequencing = init.sequencing ?? true;
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
    }
    async replay() {
        for await (const [id, signature, message] of this.iterate()) {
            if (this.signatures) {
                assert(signature !== null, "missing message signature");
                verifySignature(signature, message);
            }
            assert(message.topic === this.topic, "invalid message topic");
            assert(this.validate(message.payload));
            await this.apply(id, signature, message);
        }
    }
    async *iterate(lowerBound = null, upperBound = null, options = {}) {
        for await (const [key, value] of this.entries(AbstractMessageLog.bound(lowerBound?.id, lowerBound?.inclusive), AbstractMessageLog.bound(upperBound?.id, upperBound?.inclusive), options)) {
            const [id, signature, message] = this.decode(value);
            assert(id === decodeId(key), "expected id === decodeId(key)");
            yield [id, signature, message];
        }
    }
    encode(signature, message) {
        if (this.sequencing) {
            assert(message.clock > 0, "expected message.clock > 0 if sequencing is enable");
        }
        else {
            assert(message.clock === 0, "expected message.clock === 0 if sequencing is disabled");
        }
        if (this.signatures) {
            assert(signature !== null, "missing message signature");
        }
        return encodeSignedMessage(signature, message);
    }
    decode(value) {
        const [id, signature, message] = decodeSignedMessage(this.topic, value);
        if (this.signatures) {
            assert(signature !== null, "missing message signature");
        }
        const { topic, clock, parents, payload } = message;
        if (this.sequencing) {
            assert(clock > 0, "expected message.clock > 0 if sequencing is enabled");
        }
        else {
            assert(clock === 0, "expected message.clock === 0 if sequencing is disabled");
        }
        assert(this.validate(payload), "invalid message payload");
        return [id, signature, { topic, clock, parents, payload }];
    }
    async getClock() {
        if (this.sequencing === false) {
            return [0, []];
        }
        const parents = await this.read((txn) => this.getParents(txn));
        const clock = getClock(parents);
        return [clock, parents.map(decodeId)];
    }
    async get(id) {
        const value = await this.read(({ messages }) => messages.get(encodeId(id)));
        if (value === null) {
            return [null, null];
        }
        const [_, signature, message] = this.decode(value);
        return [signature, message];
    }
    async getParents(txn) {
        const parents = [];
        for await (const [key, value] of txn.parents.entries()) {
            assert(equals(value, cborNull), "internal error - unexpected parent entry value");
            parents.push(key);
        }
        return parents;
    }
    async append(payload, options = {}) {
        const signer = options.signer ?? AbstractMessageLog.defaultSigner;
        const { id, signature, message, result, root } = await this.write(async (txn) => {
            const parents = await this.getParents(txn);
            const clock = this.sequencing ? getClock(parents) : 0;
            const message = { topic: this.topic, clock, parents: parents.map(decodeId), payload };
            const signature = await signer.sign(message);
            const [key, value] = this.encode(signature, message);
            const id = decodeId(key);
            this.log("appending message %s", id);
            const result = await __classPrivateFieldGet(this, _AbstractMessageLog_instances, "m", _AbstractMessageLog_apply).call(this, txn, id, signature, message, [key, value]);
            const root = await txn.messages.getRoot();
            return { id, signature, message, result, root };
        });
        this.dispatchEvent(new CustomEvent("commit", { detail: { topic: this.topic, root } }));
        this.log("commited root %s", hex(root.hash));
        return { id, signature, message, result };
    }
    async insert(signature, message) {
        if (this.signatures) {
            assert(signature !== null, "missing message signature");
            verifySignature(signature, message);
        }
        const { id, root } = await this.write(async (txn) => {
            const [key, value] = this.encode(signature, message);
            const id = decodeId(key);
            this.log("inserting message %s", id);
            const dependencies = new Set();
            this.log("looking up %s parents", message.parents.length);
            for (const parentId of message.parents) {
                const parent = await txn.messages.get(encodeId(parentId));
                if (parent === null) {
                    this.log("missing parent %s", parentId);
                    dependencies.add(parentId);
                }
                else {
                    this.log("found parent %s", parentId);
                }
            }
            if (dependencies.size > 0) {
                this.log("missing %d/%d parents", dependencies.size, message.parents.length);
                this.log("mempool.messages.set(%s, ...)", id);
                this.mempool.messages.set(id, { signature, message });
                this.log("mempool.dependencies.set(%s, %o)", id, dependencies);
                this.mempool.dependencies.set(id, dependencies);
                for (const parent of dependencies) {
                    this.log("mempool.children[%s].add(%s)", parent, id);
                    const children = this.mempool.children.get(parent);
                    if (children === undefined) {
                        this.mempool.children.set(parent, new Set([id]));
                    }
                    else {
                        children.add(id);
                    }
                }
                return { id };
            }
            await __classPrivateFieldGet(this, _AbstractMessageLog_instances, "m", _AbstractMessageLog_apply).call(this, txn, id, signature, message, [key, value]);
            const root = await txn.messages.getRoot();
            return { id, root };
        });
        if (root !== undefined) {
            this.dispatchEvent(new CustomEvent("commit", { detail: { topic: this.topic, root } }));
            this.log("commited root %s", hex(root.hash));
        }
        return { id };
    }
    async sync(sourcePeerId, source) {
        const root = await this.write(async (txn) => {
            const driver = new Driver(this.topic, source, txn.messages);
            for await (const [key, value] of driver.sync()) {
                const [id, signature, message] = this.decode(value);
                assert(id === decodeId(key), "expected id === decodeId(key)");
                if (this.signatures) {
                    assert(signature !== null, "missing message signature");
                    verifySignature(signature, message);
                }
                const existingMessage = await txn.messages.get(key);
                if (existingMessage === null) {
                    await __classPrivateFieldGet(this, _AbstractMessageLog_instances, "m", _AbstractMessageLog_apply).call(this, txn, id, signature, message, [key, value]);
                }
            }
            return await txn.messages.getRoot();
        }, { source: sourcePeerId });
        this.dispatchEvent(new CustomEvent("sync", { detail: { topic: this.topic, peerId: sourcePeerId } }));
        this.dispatchEvent(new CustomEvent("commit", { detail: { topic: this.topic, root } }));
        this.log("commited root %s", hex(root.hash));
        return { root };
    }
    async serve(targetPeerId, callback) {
        await this.read((txn) => callback(txn.messages), { target: targetPeerId });
    }
}
_AbstractMessageLog_instances = new WeakSet(), _AbstractMessageLog_apply = async function _AbstractMessageLog_apply(txn, id, signature, message, [key, value] = this.encode(signature, message)) {
    this.log("applying %s %O", id, message);
    const result = await this.apply(id, signature, message);
    this.dispatchEvent(new CustomEvent("message", { detail: { topic: this.topic, id, signature, message, result } }));
    await txn.messages.set(key, value);
    if (this.sequencing) {
        await txn.parents.set(key, cborNull);
        for (const parentId of message.parents) {
            await txn.parents.delete(encodeId(parentId));
        }
        const children = this.mempool.children.get(id);
        this.log("%s has %d mempool children", id, children?.size ?? 0, children);
        if (children !== undefined) {
            for (const childId of children) {
                const dependencies = this.mempool.dependencies.get(childId);
                assert(dependencies !== undefined, "expected dependencies !== undefined");
                const signedMessage = this.mempool.messages.get(childId);
                assert(signedMessage !== undefined, "expected signedMessage !== undefined");
                dependencies.delete(id);
                if (dependencies.size === 0) {
                    this.mempool.dependencies.delete(childId);
                    this.mempool.messages.delete(childId);
                    const { signature, message } = signedMessage;
                    await __classPrivateFieldGet(this, _AbstractMessageLog_instances, "m", _AbstractMessageLog_apply).call(this, txn, childId, signature, message);
                }
            }
            this.mempool.children.delete(id);
        }
    }
    return result;
};
AbstractMessageLog.defaultSigner = { sign: ({}) => null };
AbstractMessageLog.bound = (id = null, inclusive = true) => id === null ? null : { key: encodeId(id), inclusive };
class Mempool {
    constructor() {
        this.messages = new Map();
        this.dependencies = new Map();
        this.children = new Map();
    }
}
