import pDefer from "p-defer";
import { bytesToHex } from "@noble/hashes/utils";
import { IDBStore, IDBTree } from "@canvas-js/okra-idb";
import { openDB } from "idb";
import { AbstractMessageLog } from "../AbstractMessageLog.js";
import { assert } from "../utils.js";
export class MessageLog extends AbstractMessageLog {
    static async open(name, init) {
        const storeNames = [`${init.topic}/messages`, `${init.topic}/parents`];
        const db = await openDB(name, 1, {
            upgrade: (db, oldVersion, newVersion) => {
                for (const storeName of storeNames) {
                    if (db.objectStoreNames.contains(storeName)) {
                        continue;
                    }
                    else {
                        db.createObjectStore(storeName);
                    }
                }
            },
        });
        const messages = await IDBTree.open(db, `${init.topic}/messages`);
        const parents = new IDBStore(db, `${init.topic}/parents`);
        const messageLog = new MessageLog(db, messages, parents, init);
        if (init.replay) {
            await messageLog.replay();
        }
        return messageLog;
    }
    constructor(db, messages, parents, init) {
        super(init);
        this.db = db;
        this.messages = messages;
        this.parents = parents;
        this.incomingSyncPeers = new Set();
        this.outgoingSyncPeers = new Set();
        this.controller = new AbortController();
        this.lockName = bytesToHex(crypto.getRandomValues(new Uint8Array(16)));
        db.addEventListener("error", (event) => this.log("db: error", event));
        db.addEventListener("close", (event) => this.log("db: close", event));
        db.addEventListener("versionchange", (event) => {
            this.log("db: versionchange", event);
            if (event.oldVersion === null && event.newVersion !== null) {
                return;
            }
            else if (event.oldVersion !== null && event.newVersion !== null) {
                return;
            }
            else if (event.oldVersion !== null && event.newVersion === null) {
                db.close();
                return;
            }
        });
    }
    async close() {
        this.log("closing");
        this.controller.abort();
        this.db.close();
    }
    async *entries(lowerBound = null, upperBound = null, options = {}) {
        this.log("requesting shared lock");
        const deferred = pDefer();
        navigator.locks.request(this.lockName, { mode: "shared", signal: this.controller.signal }, (lock) => {
            if (lock === null) {
                this.log.error("failed to acquire shared lock");
                throw new Error(`failed to acquire shared lock ${this.lockName}`);
            }
            this.log("acquired shared lock");
            return deferred.promise;
        });
        if (this.messages.store.txn === null) {
            this.messages.store.txn = this.db.transaction([this.messages.store.storeName], "readonly");
        }
        try {
            for await (const node of this.messages.nodes(0, lowerBound ?? { key: null, inclusive: false }, upperBound, options)) {
                assert(node.key !== null, "expected node.key !== null");
                assert(node.value !== undefined, "expected node.value !== undefined");
                yield [node.key, node.value];
            }
        }
        finally {
            this.messages.store.txn = null;
            this.log("releasing shared lock");
            deferred.resolve();
        }
    }
    async read(callback, options = {}) {
        const targetPeerId = options.target ?? null;
        if (targetPeerId !== null) {
            if (this.outgoingSyncPeers.has(targetPeerId.toString())) {
                throw new Error(`deadlock with peer ${targetPeerId}`);
            }
        }
        let result = undefined;
        this.log("requesting shared lock");
        await navigator.locks.request(this.lockName, { mode: "shared", signal: this.controller.signal }, async (lock) => {
            if (lock === null) {
                this.log.error("failed to acquire shared lock");
                throw new Error(`failed to acquire shared lock ${this.lockName}`);
            }
            this.log("acquired shared lock");
            if (targetPeerId !== null) {
                this.incomingSyncPeers.add(targetPeerId.toString());
            }
            const parents = {
                get: (key) => this.parents.read(() => this.parents.get(key)),
                entries: (lowerBound = null, upperBound = null, options = {}) => {
                    this.parents.txn = this.db.transaction([this.parents.storeName], "readonly");
                    return this.parents.entries(lowerBound, upperBound, options);
                },
            };
            try {
                result = await callback({ messages: this.messages, parents });
            }
            catch (err) {
                this.log.error("error in read-only transaction: %O", err);
            }
            finally {
                this.log("releasing shared lock");
                if (targetPeerId !== null) {
                    this.incomingSyncPeers.delete(targetPeerId.toString());
                }
            }
        });
        return result;
    }
    async write(callback, options = {}) {
        const sourcePeerId = options.source ?? null;
        if (sourcePeerId !== null) {
            if (this.incomingSyncPeers.has(sourcePeerId.toString())) {
                throw new Error(`deadlock with peer ${sourcePeerId}`);
            }
        }
        let result = undefined;
        this.log("requesting exclusive lock");
        await navigator.locks.request(this.lockName, { mode: "exclusive", signal: this.controller.signal }, async (lock) => {
            if (lock === null) {
                this.log.error("failed to exclusive lock");
                throw new Error(`failed to acquire exclusive lock ${this.lockName}`);
            }
            this.log("acquired exclusive lock");
            if (sourcePeerId !== null) {
                this.outgoingSyncPeers.add(sourcePeerId.toString());
            }
            const parents = {
                get: (key) => this.parents.read(() => this.parents.get(key)),
                set: (key, value) => this.parents.write(() => this.parents.set(key, value)),
                delete: (key) => this.parents.write(() => this.parents.delete(key)),
                entries: (lowerBound = null, upperBound = null, options = {}) => {
                    this.parents.txn = this.db.transaction([`${this.topic}/parents`], "readonly");
                    return this.parents.entries(lowerBound, upperBound, options);
                },
            };
            try {
                result = await callback({ messages: this.messages, parents });
            }
            catch (err) {
                this.log.error("error in read-write transaction: %O", err);
                throw err;
            }
            finally {
                this.log("releasing exclusive lock");
                if (sourcePeerId !== null) {
                    this.outgoingSyncPeers.delete(sourcePeerId.toString());
                }
            }
        });
        return result;
    }
}
