import type { PeerId } from "@libp2p/interface-peer-id";
import type { Source, Target, Node, Bound } from "@canvas-js/okra";
import { EventEmitter } from "@libp2p/interfaces/events";
import { Logger } from "@libp2p/logger";
import type { Message } from "@canvas-js/interfaces";
import { Signature } from "@canvas-js/signed-cid";
import { Awaitable } from "../utils.js";
export interface ReadOnlyTransaction extends Target {
    get(key: Uint8Array): Awaitable<Uint8Array | null>;
    getUserdata(): Awaitable<Uint8Array | null>;
}
export interface ReadWriteTransaction extends ReadOnlyTransaction {
    set(key: Uint8Array, value: Uint8Array): Awaitable<void>;
    delete(key: Uint8Array): Awaitable<void>;
    setUserdata(userdata: Uint8Array | null): Awaitable<void>;
}
export interface MessageLogInit<Payload = unknown, Result = void> {
    location: string | null;
    topic: string;
    apply: (id: string, signature: Signature | null, message: Message<Payload>) => Awaitable<Result>;
    validate: (payload: unknown) => payload is Payload;
    signatures?: boolean;
    sequencing?: boolean;
    replay?: boolean;
}
export interface MessageSigner<Payload = unknown> {
    sign: (message: Message<Payload>) => Awaitable<Signature | null>;
}
export declare abstract class AbstractMessageLog<Payload = unknown, Result = unknown> extends EventEmitter<{
    message: CustomEvent<{
        id: string;
        signature: Signature | null;
        message: Message<Payload>;
        result: Result;
    }>;
    commit: CustomEvent<{
        root: Node;
    }>;
    sync: CustomEvent<{
        peerId: PeerId;
    }>;
}> {
    private static defaultSigner;
    private static bound;
    abstract close(): Promise<void>;
    protected abstract entries(lowerBound?: Bound<Uint8Array> | null, upperBound?: Bound<Uint8Array> | null, options?: {
        reverse?: boolean;
    }): AsyncIterable<[key: Uint8Array, value: Uint8Array]>;
    protected abstract read<T>(callback: (txn: ReadOnlyTransaction) => Promise<T>, options?: {
        target?: PeerId;
    }): Promise<T>;
    protected abstract write<T>(callback: (txn: ReadWriteTransaction) => Promise<T>, options?: {
        source?: PeerId;
    }): Promise<T>;
    readonly location: string | null;
    readonly topic: string;
    readonly apply: (id: string, signature: Signature | null, message: Message<Payload>) => Awaitable<Result>;
    readonly validate: (payload: unknown) => payload is Payload;
    readonly signatures: boolean;
    readonly sequencing: boolean;
    protected readonly log: Logger;
    protected constructor(init: MessageLogInit<Payload, Result>);
    iterate(lowerBound?: {
        id: string;
        inclusive: boolean;
    } | null, upperBound?: {
        id: string;
        inclusive: boolean;
    } | null, options?: {
        reverse?: boolean;
    }): AsyncIterable<[id: string, signature: Signature | null, message: Message<Payload>]>;
    encode(signature: Signature | null, message: Message<Payload>): [key: Uint8Array, value: Uint8Array];
    decode(value: Uint8Array): [key: Uint8Array, signature: Signature | null, message: Message<Payload>];
    getClock(): Promise<[clock: number, parents: string[]]>;
    get(id: string): Promise<[signature: Signature | null, message: Message<Payload> | null]>;
    insert(signature: Signature | null, message: Message<Payload>): Promise<{
        id: string;
        key: Uint8Array;
        value: Uint8Array;
        result: Awaited<Result>;
        root: Node;
    }>;
    append(payload: Payload, options?: {
        signer?: MessageSigner<Payload>;
    }): Promise<{
        id: string;
        key: Uint8Array;
        value: Uint8Array;
        result: Awaited<Result>;
        root: Node;
    }>;
    sync(sourcePeerId: PeerId, source: Source): Promise<{
        root: Node;
    }>;
    serve(targetPeerId: PeerId, callback: (source: Source) => Promise<void>): Promise<void>;
}
