var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Graph_parents;
import { CustomEvent, EventEmitter } from "@libp2p/interfaces/events";
import { logger } from "@libp2p/logger";
import { equals } from "uint8arrays";
import * as cbor from "@ipld/dag-cbor";
import { verifySignature } from "@canvas-js/signed-cid";
import { KEY_LENGTH, decodeId, decodeSignedMessage, encodeId, encodeSignedMessage, getClock } from "../schema.js";
import { Driver } from "../sync/driver.js";
import { assert, nsidPattern } from "../utils.js";
export class AbstractMessageLog extends EventEmitter {
    constructor(init) {
        super();
        assert(nsidPattern.test(init.topic), "invalid topic (must match [a-zA-Z0-9\\.\\-]+)");
        this.location = init.location;
        this.topic = init.topic;
        this.apply = init.apply;
        this.validate = init.validate;
        this.signatures = init.signatures ?? true;
        this.sequencing = init.sequencing ?? true;
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
    }
    async *iterate(lowerBound = null, upperBound = null, options = {}) {
        for await (const [key, value] of this.entries(AbstractMessageLog.bound(lowerBound?.id, lowerBound?.inclusive), AbstractMessageLog.bound(upperBound?.id, upperBound?.inclusive), options)) {
            const [recoveredKey, signature, message] = this.decode(value);
            assert(equals(recoveredKey, key), "expected equals(recoveredKey, key)");
            yield [decodeId(key), signature, message];
        }
    }
    encode(signature, message) {
        if (this.sequencing) {
            assert(message.clock > 0, "expected message.clock > 0 if sequencing is enable");
        }
        else {
            assert(message.clock === 0, "expected message.clock === 0 if sequencing is disabled");
        }
        if (this.signatures) {
            assert(signature !== null, "missing message signature");
        }
        return encodeSignedMessage(signature, message);
    }
    decode(value) {
        const [key, signature, message] = decodeSignedMessage(value);
        if (this.signatures) {
            assert(signature !== null, "missing message signature");
        }
        const { clock, parents, payload } = message;
        if (this.sequencing) {
            assert(clock > 0, "expected message.clock > 0 if sequencing is enable");
        }
        else {
            assert(clock === 0, "expected message.clock === 0 if sequencing is disabled");
        }
        assert(this.validate(payload), "invalid message payload");
        return [key, signature, { clock, parents, payload }];
    }
    async getClock() {
        return await this.read(async (txn) => {
            const userdata = await txn.getUserdata();
            const graph = Graph.import(this.topic, this.sequencing, userdata);
            const [clock, parents] = graph.getClock();
            return [clock, parents];
        });
    }
    async get(id) {
        const key = encodeId(id);
        assert(key.length === KEY_LENGTH, "invalid id");
        const value = await this.read(async (txn) => txn.get(key));
        if (value === null) {
            return [null, null];
        }
        const [recoveredKey, signature, message] = this.decode(value);
        assert(equals(recoveredKey, key), "invalid message key");
        return [signature, message];
    }
    async insert(signature, message) {
        return await this.write(async (txn) => {
            const oldUserdata = await txn.getUserdata();
            const graph = Graph.import(this.topic, this.sequencing, oldUserdata);
            const [key, value] = this.encode(signature, message);
            const id = decodeId(key);
            this.log("inserting message %s at clock %d with %d parents", id, message.clock, message.parents.length);
            const result = await this.apply(id, signature, message);
            this.dispatchEvent(new CustomEvent("message", { detail: { id, signature, message, result } }));
            await txn.set(key, value);
            graph.update(id, message);
            const newUserdata = graph.export();
            if (newUserdata !== null) {
                await txn.setUserdata(newUserdata);
            }
            const root = await txn.getRoot();
            return { id, key, value, result, root };
        });
    }
    async append(payload, options = {}) {
        const signer = options.signer ?? AbstractMessageLog.defaultSigner;
        return await this.write(async (txn) => {
            const oldUserdata = await txn.getUserdata();
            const graph = Graph.import(this.topic, this.sequencing, oldUserdata);
            const message = graph.create(payload);
            const signature = await signer.sign(message);
            const [key, value] = this.encode(signature, message);
            const id = decodeId(key);
            this.log("appending message %s at clock %d with %d parents", id, message.clock, message.parents.length);
            const result = await this.apply(id, signature, message);
            this.dispatchEvent(new CustomEvent("message", { detail: { id, signature, message, result } }));
            await txn.set(key, value);
            graph.update(id, message);
            const newUserdata = graph.export();
            if (newUserdata !== null) {
                await txn.setUserdata(newUserdata);
            }
            const root = await txn.getRoot();
            return { id, key, value, result, root };
        });
    }
    async sync(sourcePeerId, source) {
        const root = await this.write(async (target) => {
            const oldUserdata = await target.getUserdata();
            const graph = Graph.import(this.topic, this.sequencing, oldUserdata);
            const driver = new Driver(this.topic, source, target);
            for await (const [key, value] of driver.sync()) {
                const id = decodeId(key);
                const [recoveredKey, signature, message] = this.decode(value);
                assert(equals(key, recoveredKey), "invalid message key");
                if (this.signatures) {
                    assert(signature !== null, "missing message signature");
                    verifySignature(signature, message);
                }
                this.log("inserting message %s at clock %d with %d parents", id, message.clock, message.parents.length);
                const result = await this.apply(id, signature, message);
                this.dispatchEvent(new CustomEvent("message", { detail: { id, signature, message, result } }));
                await target.set(key, value);
                graph.update(id, message);
            }
            const newUserdata = graph.export();
            if (newUserdata !== null) {
                await target.setUserdata(newUserdata);
            }
            return await target.getRoot();
        }, { source: sourcePeerId });
        this.dispatchEvent(new CustomEvent("commit", { detail: { root } }));
        this.dispatchEvent(new CustomEvent("sync", { detail: { peerId: sourcePeerId } }));
        return { root };
    }
    async serve(targetPeerId, callback) {
        await this.read(callback, { target: targetPeerId });
    }
}
AbstractMessageLog.defaultSigner = { sign: ({}) => null };
AbstractMessageLog.bound = (id = null, inclusive = true) => id === null ? null : { key: encodeId(id), inclusive };
class Graph {
    static import(topic, sequencing, userdata) {
        if (sequencing === false) {
            return new Graph(topic, sequencing, []);
        }
        const parents = userdata === null ? [] : cbor.decode(userdata);
        return new Graph(topic, sequencing, parents);
    }
    constructor(topic, sequencing, parents) {
        this.sequencing = sequencing;
        _Graph_parents.set(this, new Set());
        const clock = getClock(parents);
        this.log = logger(`canvas:gossiplog:[${topic}]:graph`);
        this.log("loaded graph at clock %d", clock);
        for (const key of parents) {
            assert(key.byteLength === KEY_LENGTH, "expected key.byteLength === KEY_LENGTH");
            const id = decodeId(key);
            this.log("added %s", id);
            __classPrivateFieldGet(this, _Graph_parents, "f").add(id);
        }
    }
    export() {
        if (this.sequencing === false) {
            return null;
        }
        const [clock, parents, keys] = this.getClock();
        this.log("saving graph at clock %d", clock);
        return cbor.encode(keys);
    }
    update(id, message) {
        if (this.sequencing === false) {
            return;
        }
        for (const parent of message.parents) {
            if (__classPrivateFieldGet(this, _Graph_parents, "f").delete(parent)) {
                this.log("removed %s", parent);
            }
        }
        this.log("added %s", id);
        __classPrivateFieldGet(this, _Graph_parents, "f").add(id);
    }
    create(payload) {
        if (this.sequencing === false) {
            return { clock: 0, parents: [], payload };
        }
        const [clock, parents] = this.getClock();
        return { clock, parents, payload };
    }
    getClock() {
        if (this.sequencing === false) {
            return [0, [], []];
        }
        const ids = Array.from(__classPrivateFieldGet(this, _Graph_parents, "f")).sort();
        const keys = ids.map(encodeId);
        const clock = getClock(keys);
        return [clock, ids, keys];
    }
}
_Graph_parents = new WeakMap();
