var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Graph_parents;
import { logger } from "@libp2p/logger";
import * as cbor from "@ipld/dag-cbor";
import { base32 } from "multiformats/bases/base32";
import { equals } from "uint8arrays";
import { verifySignature } from "@canvas-js/signed-cid";
import { Driver } from "../sync/driver.js";
import { decodeSignedMessage, getClock } from "../schema.js";
import { assert } from "../utils.js";
export class AbstractStore {
    constructor(init) {
        this.init = init;
        this.log = logger(`canvas:gossiplog:store`);
    }
    async get(key) {
        const value = await this.read(async (txn) => txn.get(key));
        if (value === null) {
            return [null, null];
        }
        const [recoveredKey, signature, message] = decodeSignedMessage(value, {
            sequencing: this.init.sequencing,
            signatures: this.init.signatures,
        });
        assert(equals(recoveredKey, key), "invalid message key");
        return [signature, message];
    }
    async sync(peerId, source, callback) {
        let root = null;
        await this.target(peerId, async (target) => {
            const graph = await Graph.import(target);
            const driver = new Driver(source, target);
            for await (const [key, value] of driver.sync()) {
                const id = base32.baseEncode(key);
                let signedMessage = null;
                try {
                    const [recoveredKey, signature, message] = decodeSignedMessage(value, {
                        sequencing: this.init.sequencing,
                        signatures: this.init.signatures,
                    });
                    assert(equals(key, recoveredKey), "invalid message key");
                    signedMessage = { signature, message };
                }
                catch (err) {
                    this.log.error("[%s] failed to decode signed message %s: %O", this.init.topic, id, err);
                    continue;
                }
                const { signature, message } = signedMessage;
                if (this.init.signatures) {
                    try {
                        assert(signature !== null, "missing message signature");
                        verifySignature(signature, message);
                    }
                    catch (err) {
                        this.log.error("[%s] invalid signature for message %s: %O", this.init.topic, id, err);
                        continue;
                    }
                }
                try {
                    await callback(key, signature, message);
                }
                catch (err) {
                    this.log.error("[%s] failed to apply message %s: %O", this.init.topic, id, err);
                    continue;
                }
                await target.set(key, value);
                graph.update(key, message);
            }
            await graph.save(target);
            root = await target.getRoot();
        });
        assert(root !== null, "internal error - sync exited prematurely");
        return { root };
    }
}
export class Graph {
    static async import(txn) {
        const userdata = await txn.getUserdata();
        const parents = userdata === null ? [] : cbor.decode(userdata);
        return new Graph(parents);
    }
    constructor(parents) {
        this.log = logger("canvas:gossiplog:graph");
        _Graph_parents.set(this, new Set());
        const clock = getClock(parents);
        this.log("created graph at clock %d with parents %o", clock, parents.map(base32.baseEncode));
        for (const key of parents) {
            __classPrivateFieldGet(this, _Graph_parents, "f").add(base32.baseEncode(key));
        }
    }
    async save(txn) {
        const [clock, parents] = this.export();
        this.log("saving graph at clock %d with parents %o", clock, parents.map(base32.baseEncode));
        await txn.setUserdata(cbor.encode(parents));
    }
    update(key, message) {
        for (const parent of message.parents) {
            const parentId = base32.baseEncode(parent);
            if (__classPrivateFieldGet(this, _Graph_parents, "f").delete(parentId)) {
                this.log("removed %s", parentId);
            }
        }
        const id = base32.baseEncode(key);
        this.log("added %s", id);
        __classPrivateFieldGet(this, _Graph_parents, "f").add(id);
    }
    export() {
        const parents = [...__classPrivateFieldGet(this, _Graph_parents, "f")].map(base32.baseDecode);
        const clock = getClock(parents);
        return [clock, parents];
    }
}
_Graph_parents = new WeakMap();
