import pDefer from "p-defer";
import { bytesToHex } from "@noble/hashes/utils";
import { openDB } from "idb";
import { IDBTree } from "@canvas-js/okra-idb";
import { assert } from "@canvas-js/okra";
import openMemoryMessageLog from "../memory/index.js";
import { AbstractMessageLog } from "../AbstractMessageLog.js";
export * from "../AbstractMessageLog.js";
export default async function openMessageLog(init) {
    if (init.location === null) {
        return openMemoryMessageLog(init);
    }
    const storeNames = [init.topic];
    const db = await openDB(init.location, 1, {
        upgrade: (db, oldVersion, newVersion) => {
            for (const storeName of storeNames) {
                if (db.objectStoreNames.contains(storeName)) {
                    continue;
                }
                else {
                    db.createObjectStore(storeName);
                }
            }
        },
    });
    const tree = await IDBTree.open(db, init.topic);
    return new Messagelog(init, db, tree);
}
class Messagelog extends AbstractMessageLog {
    constructor(init, db, tree) {
        super(init);
        this.db = db;
        this.tree = tree;
        this.incomingSyncPeers = new Set();
        this.outgoingSyncPeers = new Set();
        this.controller = new AbortController();
        this.lockName = bytesToHex(crypto.getRandomValues(new Uint8Array(16)));
    }
    async close() {
        this.controller.abort();
        this.db.close();
    }
    async *entries(lowerBound = null, upperBound = null, options = {}) {
        this.log("requesting shared lock");
        const deferred = pDefer();
        navigator.locks.request(this.lockName, { mode: "shared", signal: this.controller.signal }, (lock) => {
            if (lock === null) {
                this.log.error("failed to acquire shared lock");
                throw new Error(`failed to acquire shared lock ${this.lockName}`);
            }
            this.log("acquired shared lock");
            return deferred.promise;
        });
        try {
            for await (const node of this.tree.nodes(0, lowerBound ?? { key: null, inclusive: false }, upperBound, options)) {
                assert(node.key !== null, "expected node.key !== null");
                assert(node.value !== undefined, "expected node.value !== undefined");
                yield [node.key, node.value];
            }
        }
        finally {
            this.log("releasing shared lock");
            deferred.resolve();
        }
    }
    async read(callback, options = {}) {
        const targetPeerId = options.target ?? null;
        if (targetPeerId !== null) {
            if (this.outgoingSyncPeers.has(targetPeerId.toString())) {
                throw new Error(`deadlock with peer ${targetPeerId}`);
            }
        }
        let result = undefined;
        this.log("requesting shared lock");
        await navigator.locks.request(this.lockName, { mode: "shared", signal: this.controller.signal }, async (lock) => {
            if (lock === null) {
                this.log.error("failed to acquire shared lock");
                throw new Error(`failed to acquire shared lock ${this.lockName}`);
            }
            this.log("acquired shared lock");
            if (targetPeerId !== null) {
                this.incomingSyncPeers.add(targetPeerId.toString());
            }
            try {
                result = await callback(this.tree);
            }
            catch (err) {
                this.log.error("error in read-only transaction: %O", err);
            }
            finally {
                this.log("releasing shared lock");
                if (targetPeerId !== null) {
                    this.incomingSyncPeers.delete(targetPeerId.toString());
                }
            }
        });
        return result;
    }
    async write(callback, options = {}) {
        const sourcePeerId = options.source ?? null;
        if (sourcePeerId !== null) {
            if (this.incomingSyncPeers.has(sourcePeerId.toString())) {
                throw new Error(`deadlock with peer ${sourcePeerId}`);
            }
        }
        let result = undefined;
        this.log("requesting exclusive lock");
        await navigator.locks.request(this.lockName, { mode: "exclusive", signal: this.controller.signal }, async (lock) => {
            if (lock === null) {
                this.log.error("failed to exclusive lock");
                throw new Error(`failed to acquire exclusive lock ${this.lockName}`);
            }
            this.log("acquired exclusive lock");
            if (sourcePeerId !== null) {
                this.outgoingSyncPeers.add(sourcePeerId.toString());
            }
            try {
                result = await callback(this.tree);
            }
            catch (err) {
                this.log.error("error in read-write transaction: %O", err);
                throw err;
            }
            finally {
                this.log("releasing exclusive lock");
                if (sourcePeerId !== null) {
                    this.outgoingSyncPeers.delete(sourcePeerId.toString());
                }
            }
        });
        return result;
    }
}
