import PQueue from "p-queue";
import pDefer from "p-defer";
import { MemoryTree } from "@canvas-js/okra-memory";
import { assert } from "@canvas-js/okra";
import { AbstractMessageLog } from "../AbstractMessageLog.js";
export default async function openMessageLog(init) {
    const tree = await MemoryTree.open();
    return new MessageLog(init, tree);
}
class MessageLog extends AbstractMessageLog {
    constructor(init, tree) {
        super(init);
        this.tree = tree;
        this.queue = new PQueue({ concurrency: 1 });
        this.incomingSyncPeers = new Set();
        this.outgoingSyncPeers = new Set();
    }
    async close() { }
    async *entries(lowerBound = null, upperBound = null, options = {}) {
        const deferred = pDefer();
        this.log("adding transaction to queue");
        this.queue.add(() => {
            this.log("beginning transaction");
            return deferred.promise;
        });
        try {
            for await (const node of this.tree.nodes(0, lowerBound ?? { key: null, inclusive: false }, upperBound, options)) {
                assert(node.key !== null, "expected node.key !== null");
                assert(node.value !== undefined, "expected node.value !== undefined");
                yield [node.key, node.value];
            }
        }
        finally {
            this.log("transaction completed");
            deferred.resolve();
        }
    }
    async read(callback, options = {}) {
        const targetPeerId = options.target ?? null;
        if (targetPeerId !== null) {
            if (this.outgoingSyncPeers.has(targetPeerId.toString())) {
                throw new Error(`deadlock with peer ${targetPeerId}`);
            }
        }
        this.log("adding transaction to queue");
        const result = await this.queue.add(async () => {
            this.log("beginning transaction");
            if (targetPeerId !== null) {
                this.incomingSyncPeers.add(targetPeerId.toString());
            }
            try {
                return await callback(this.tree);
            }
            catch (err) {
                this.log.error("error in transaction: %O", err);
            }
            finally {
                this.log("transaction completed");
                if (targetPeerId !== null) {
                    this.incomingSyncPeers.delete(targetPeerId.toString());
                }
            }
        });
        return result;
    }
    async write(callback, options = {}) {
        const sourcePeerId = options.source ?? null;
        if (sourcePeerId !== null) {
            if (this.incomingSyncPeers.has(sourcePeerId.toString())) {
                throw new Error(`deadlock with peer ${sourcePeerId}`);
            }
        }
        this.log("adding transaction to queue");
        const result = await this.queue.add(async () => {
            this.log("beginning transaction");
            if (sourcePeerId !== null) {
                this.outgoingSyncPeers.add(sourcePeerId.toString());
            }
            try {
                return await callback(this.tree);
            }
            catch (err) {
                this.log.error("error in transaction: %O", err);
                throw err;
            }
            finally {
                this.log("transaction completed");
                if (sourcePeerId !== null) {
                    this.outgoingSyncPeers.delete(sourcePeerId.toString());
                }
            }
        });
        return result;
    }
}
